@@ -38,6 +38,7 @@
 #include "config.h"
 #endif
 
+#include <assert.h>
 #include <sys/time.h>
 #include <picturestr.h>
 #include <psb_reg.h>
@@ -143,6 +144,9 @@ psbPixelARGB8888(unsigned format, void *pixelP, CARD32 * argb8888)
 {
     CARD32 bits, shift, pixel, bpp;
 
+    if (pixelP == NULL)
+	    return;
+
     bpp = PICT_FORMAT_BPP(format);
 
     if (bpp <= 8) {
@@ -310,8 +314,10 @@ psbExaPixmapIsOffscreen(PixmapPtr p)
     ScreenPtr pScreen = p->drawable.pScreen;
     ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
     PsbPtr pPsb = psbPTR(pScrn);
-
-    return (psbInBuffer(&pPsb->buffers, p->devPrivate.ptr) != NULL);
+    
+    return ((unsigned long)p->devPrivate.ptr 
+		- (unsigned long)mmBufVirtual(pPsb->pPsbExa->exaBuf.buf)) 
+			< mmBufSize(pPsb->pPsbExa->exaBuf.buf);
 }
 
 void
@@ -343,6 +349,10 @@ psbExaAllocBuffers(ScrnInfoPtr pScrn, PsbExaPtr pPsbExa)
     PsbPtr pPsb = psbPTR(pScrn);
     PsbDevicePtr pDevice = psbDevicePTR(pPsb);
 
+#ifdef XF86DRI
+    PsbDRIPtr pPsbDRI;
+#endif
+
     mmInitListHead(&pPsbExa->exaBuf.head);
     mmInitListHead(&pPsbExa->scratchBuf.head);
     mmInitListHead(&pPsbExa->tmpBuf.head);
@@ -352,19 +362,25 @@ psbExaAllocBuffers(ScrnInfoPtr pScrn, PsbExaPtr pPsbExa)
 					  MM_FLAG_READ |
 					  MM_FLAG_WRITE |
 					  MM_FLAG_MEM_TT |
-					  ((pPsb->
-					    exaCached) ? MM_FLAG_CACHED : 0),
+ 					  MM_FLAG_SHAREABLE, /* for psbTexOffsetStart */
 					  MM_HINT_DONT_FENCE));
     if (!pPsbExa->exaBuf.buf)
 	return FALSE;
 
+#ifdef XF86DRI
+    pPsbDRI = (PsbDRIPtr) pPsb->pDRIInfo->devPrivate;
+    if (pPsbDRI) {
+        drmBO *bo = mmKernelBuf(pPsbExa->exaBuf.buf);
+        if(bo) {
+           pPsbDRI->exaBufHandle = bo->handle;
+        }
+    }
+#endif
     psbAddBufItem(&pPsb->buffers, &pPsbExa->scratchBuf,
 		  pDevice->man->createBuf(pDevice->man, pPsb->exaScratchSize,
 					  0,
 					  MM_FLAG_READ | MM_FLAG_WRITE |
-					  MM_FLAG_MEM_TT | ((pPsb->exaCached)
-							    ? MM_FLAG_CACHED :
-							    0),
+					  MM_FLAG_MEM_TT,
 					  MM_HINT_DONT_FENCE));
 
     if (!pPsbExa->scratchBuf.buf)
@@ -1187,139 +1203,6 @@ psbExaUploadToScreen(PixmapPtr pDst, int x, int y, int w, int h, char *src,
     return TRUE;
 }
 
-/*
- * Create a user ttm buffer containing the destination. Blit to it
- * and take it out.
- */
-
-static Bool
-psbAccelDownloadFromScreen(PixmapPtr pSrc, int x, int y, int w, int h,
-			   char *dst, int dst_pitch)
-{
-    ScrnInfoPtr pScrn = xf86Screens[pSrc->drawable.pScreen->myNum];
-    PsbPtr pPsb = psbPTR(pScrn);
-    PsbTwodContextPtr tdc = &pPsb->td;
-    Psb2DBufferPtr cb2 = &pPsb->superC;
-    PsbDevicePtr pDevice = psbDevicePTR(pPsb);
-    unsigned int bitsPerPixel = pSrc->drawable.bitsPerPixel;
-    CARD32 rop;
-
-    /* can't do depth 4 */
-    if (pSrc->drawable.depth == 4)
-	return FALSE;
-
-    /*
-     * Small download. Use CPU.
-     */
-
-    if (w * h * (bitsPerPixel >> 3) < PSB_EXA_MIN_DOWNLOAD)
-	return FALSE;
-
-    if (!psbExaGetSuperOffset(pSrc, &tdc->sOffset, &tdc->sBuffer))
-	return FALSE;
-
-    /*
-     * If the src buffer is cache-coherent,
-     * we do a software copy instead.
-     */
-
-    if (mmBufFlags(tdc->sBuffer) & MM_FLAG_CACHED)
-	return FALSE;
-
-    tdc->dBuffer = pDevice->man->createUserBuf(pDevice->man,
-					       dst,
-					       dst_pitch * h,
-					       MM_FLAG_MEM_TT |
-					       MM_FLAG_WRITE |
-					       MM_FLAG_CACHED |
-					       MM_FLAG_FORCE_CACHING, 0);
-    if (!tdc->dBuffer)
-	return FALSE;
-
-    /*
-     * User buf offsets always land on page boundaries.
-     */
-
-    tdc->dOffset = ((unsigned long)dst) & 0xfff;
-    tdc->direction = psbAccelCopyDirection(1, 1);
-    rop = psbCopyROP[GXcopy];
-    psbAccelSetMode(tdc, pSrc->drawable.depth, pSrc->drawable.depth,
-		    0xFFFFFFFF);
-    tdc->cmd =
-	PSB_2D_BLIT_BH | PSB_2D_ROT_NONE |
-	tdc->direction |
-	PSB_2D_DSTCK_DISABLE |
-	PSB_2D_SRCCK_DISABLE |
-	PSB_2D_USE_FILL |
-	((rop << PSB_2D_ROP3B_SHIFT) & PSB_2D_ROP3B_MASK) |
-	((rop << PSB_2D_ROP3A_SHIFT) & PSB_2D_ROP3A_MASK);
-
-    tdc->sStride = exaGetPixmapPitch(pSrc);
-    tdc->dStride = dst_pitch;
-
-    tdc->srcState = TRUE;
-    tdc->dstState = TRUE;
-    psbAccelSuperEmitState(&pPsb->superC, tdc);
-
-    psbAccelSuperCopyHelper(cb2, tdc, x, y, 0, 0, w, h,
-			    tdc->sMode, tdc->dMode, tdc->fixPat, tdc->cmd);
-
-    psbFlush2D(cb2, DRM_FENCE_FLAG_NO_USER, 0);
-
-    /*
-     * Map to sync. Really should have a buffer idle.
-     */
-
-    if (pDevice->man->mapBuf(tdc->dBuffer, DRM_BO_FLAG_READ, 0)) {
-	mmBufDestroy(tdc->dBuffer);
-	return FALSE;
-    }
-
-    pDevice->man->unMapBuf(tdc->dBuffer);
-    mmBufDestroy(tdc->dBuffer);
-    return TRUE;
-}
-
-static Bool
-psbExaDownloadFromScreen(PixmapPtr pSrc, int x, int y, int w, int h,
-			 char *dst, int dst_pitch)
-{
-    ScreenPtr pScreen = pSrc->drawable.pScreen;
-    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
-    PsbPtr pPsb = psbPTR(pScrn);
-    CARD8 *ptr;
-    PsbBufListPtr b;
-    unsigned srcPitch = exaGetPixmapPitch(pSrc);
-    unsigned int bitsPerPixel = pSrc->drawable.bitsPerPixel;
-    unsigned wBytes = (pSrc->drawable.bitsPerPixel * w) >> 3;
-
-    if (bitsPerPixel != 8 && bitsPerPixel != 16 && bitsPerPixel != 32)
-	return FALSE;
-
-    if (psbAccelDownloadFromScreen(pSrc, x, y, w, h, dst, dst_pitch))
-	return TRUE;
-
-    ptr = (CARD8 *) (exaGetPixmapOffset(pSrc) +
-		     (unsigned long)mmBufVirtual(pPsb->pPsbExa->exaBuf.buf));
-    b = psbInBuffer(&pPsb->buffers, ptr);
-    if (!b)
-	return FALSE;
-
-    ptr += y * srcPitch + ((x * bitsPerPixel) >> 3);
-
-    if (b->buf->man->mapBuf(b->buf, MM_FLAG_READ, 0))
-	return FALSE;
-
-    while (h--) {
-	memcpy(dst, ptr, wBytes);
-	dst += dst_pitch;
-	ptr += srcPitch;
-    }
-
-    b->buf->man->unMapBuf(b->buf);
-    return TRUE;
-}
-
 static void
 psbExaSuperComposite(PixmapPtr pDst, int srcX, int srcY, int maskX, int maskY,
 		     int dstX, int dstY, int width, int height)
@@ -1429,8 +1312,8 @@ psbExaInit(ScrnInfoPtr pScrn)
     pExa->pixmapOffsetAlign = 8;
     pExa->pixmapPitchAlign = 32 * 4;
     pExa->flags = EXA_OFFSCREEN_PIXMAPS;
-    pExa->maxX = 4095;
-    pExa->maxY = 4095;
+    pExa->maxX = 2047;
+    pExa->maxY = 2047;
     pExa->WaitMarker = psbExaWaitMarker;
     pExa->MarkSync = psbExaMarkSync;
     pExa->PrepareSolid = psbExaPrepareSuperSolid;
@@ -1447,7 +1330,6 @@ psbExaInit(ScrnInfoPtr pScrn)
     pExa->PrepareAccess = psbExaPrepareAccess;
     pExa->FinishAccess = psbExaFinishAccess;
     pExa->UploadToScreen = psbExaUploadToScreen;
-    pExa->DownloadFromScreen = psbExaDownloadFromScreen;
 
     if (!exaDriverInit(pScrn->pScreen, pExa)) {
 	goto out_err;
@@ -1469,3 +1351,30 @@ psbExaInit(ScrnInfoPtr pScrn)
 
     return NULL;
 }
+
+#ifdef XF86DRI
+
+#ifndef ExaOffscreenMarkUsed
+extern void ExaOffscreenMarkUsed(PixmapPtr);
+#endif
+
+#ifndef ExaOffscreenMarkUsed
+extern Bool exaPixmapIsOffscreen(PixmapPtr);
+#endif
+
+unsigned long long
+psbTexOffsetStart(PixmapPtr pPix)
+{
+    ScreenPtr pScreen = pPix->drawable.pScreen;
+    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
+
+    exaMoveInPixmap(pPix);
+    ExaOffscreenMarkUsed(pPix);
+
+    if (!exaPixmapIsOffscreen(pPix))
+        return ~0ULL;
+
+    return exaGetPixmapOffset(pPix);
+}
+
+#endif
