@@ -174,6 +174,14 @@ psb_crtc_set_cursor_position(xf86CrtcPtr crtc, int x, int y)
 	y = -y;
     }
 
+    /* downscaling only applies to pipe B */
+    if (intel_crtc->downscale && intel_crtc->pipe == 1) {
+	    if (intel_crtc->scale_x > 1) 
+		    x = x / intel_crtc->scale_x;
+	    if (intel_crtc->scale_y > 1) 
+		    y = y / intel_crtc->scale_y;
+    }
+
     temp |= ((x & CURSOR_POS_MASK) << CURSOR_X_SHIFT);
     temp |= ((y & CURSOR_POS_MASK) << CURSOR_Y_SHIFT);
 
