@@ -36,7 +36,6 @@
 
 #include <string.h>
 #include <unistd.h>
-
 #include <psb_reg.h>
 #include "psb_driver.h"
 #include "i810_reg.h"
@@ -92,6 +91,18 @@ static int psbEntityIndex = -1;
 #define makedev(x,y)    ((dev_t)(((x) << 8) | (y)))
 #endif
 
+#if (XORG_VERSION_MINOR >= 5 || (XORG_VERSION_MINOR == 4 && XORG_VERSION_PATCH == 99))
+#ifndef PCI_DOM_MASK
+#define PCI_DOM_MASK 0x0ffu
+#endif
+#define PCI_DOMBUS_MASK (((PCI_DOM_MASK) << 8)| 0x0ffu)
+
+#define PCI_DOM_FROM_TAG(tag) (((tag)>>24) & (PCI_DOM_MASK))
+#define PCI_BUS_FROM_TAG(tag) (((tag) >> 16) & (PCI_DOMBUS_MASK))
+#define PCI_DEV_FROM_TAG(tag) (((tag) & 0x0000f800u) >> 11)
+#define PCI_FUNC_FROM_TAG(tag) (((tag) & 0x00000700u) >> 8)
+#endif
+
 /*
  * This contains the functions needed by the server after loading the
  * driver module.  It must be supplied, and gets added the driver list by
@@ -134,11 +145,12 @@ typedef enum
     OPTION_SWCURSOR,
     OPTION_EXAMEM,
     OPTION_EXASCRATCH,
-    OPTION_EXACACHED,
     OPTION_IGNORE_ACPI,
     OPTION_NOPANEL,
     OPTION_LIDTIMER,
-    OPTION_NOFITTING
+    OPTION_NOFITTING,
+    OPTION_DOWNSCALE,
+    OPTION_VSYNC
 } psbOpts;
 
 static const OptionInfoRec psbOptions[] = {
@@ -147,11 +159,12 @@ static const OptionInfoRec psbOptions[] = {
     {OPTION_SWCURSOR, "SWcursor", OPTV_BOOLEAN, {0}, FALSE},
     {OPTION_EXAMEM, "ExaMem", OPTV_INTEGER, {0}, FALSE},
     {OPTION_EXASCRATCH, "ExaScratch", OPTV_INTEGER, {0}, FALSE},
-    {OPTION_EXACACHED, "ExaCached", OPTV_BOOLEAN, {0}, FALSE},
     {OPTION_IGNORE_ACPI, "IgnoreACPI", OPTV_BOOLEAN, {0}, FALSE},
     {OPTION_NOPANEL, "NoPanel", OPTV_BOOLEAN, {0}, FALSE},
     {OPTION_LIDTIMER, "LidTimer", OPTV_BOOLEAN, {0}, FALSE},
     {OPTION_NOFITTING, "NoFitting", OPTV_BOOLEAN, {0}, FALSE},
+    {OPTION_DOWNSCALE, "DownScale", OPTV_BOOLEAN, {0}, FALSE},
+    {OPTION_VSYNC, "Vsync", OPTV_BOOLEAN, {0}, FALSE},
     {-1, NULL, OPTV_NONE, {0}, FALSE},
 };
 
@@ -373,7 +386,9 @@ psbProbe(DriverPtr drv, int flags)
     if ((numDevSections = xf86MatchDevice(PSB_NAME, &devSections)) <= 0)
 	return (FALSE);
 
+#if (!(XORG_VERSION_MINOR >= 5 || (XORG_VERSION_MINOR == 4 && XORG_VERSION_PATCH == 99)))
     if (xf86GetPciVideoInfo()) {
+#endif
 
 	/*
 	 * This function allocates screens to devices according to
@@ -385,7 +400,10 @@ psbProbe(DriverPtr drv, int flags)
 					psbChipsets, psbPCIchipsets,
 					devSections, numDevSections, drv,
 					&usedChips);
+
+#if (!(XORG_VERSION_MINOR >= 5 || (XORG_VERSION_MINOR == 4 && XORG_VERSION_PATCH == 99)))
     }
+#endif
 
     if (numUsed <= 0)
 	goto out;
@@ -438,9 +456,16 @@ psbProbe(DriverPtr drv, int flags)
 	    pPsbDev->lastInstance = -1;
 	    pPsbDev->pEnt = pEnt;
 	    pPsbDev->pciInfo = xf86GetPciInfoForEntity(pEnt->index);
+           
+            #if (XORG_VERSION_MINOR >= 5 || (XORG_VERSION_MINOR == 4 && XORG_VERSION_PATCH == 99))
 	    pPsbDev->pciTag = pciTag(pPsbDev->pciInfo->bus,
-				     pPsbDev->pciInfo->device,
+				     pPsbDev->pciInfo->dev,
 				     pPsbDev->pciInfo->func);
+            #else
+	    pPsbDev->pciTag = pciTag(pPsbDev->pciInfo->bus,
+                                     pPsbDev->pciInfo->device,
+                                     pPsbDev->pciInfo->func);
+            #endif
 	    pPsbDev->refCount = 0;
 	    pPsbDev->pScrns[0] = pScrn;
 	} else {
@@ -679,16 +704,6 @@ psbPreInitAccel(ScrnInfoPtr pScrn)
 		   "[EXA] Allocate %d kiB for scratch memory.\n", tmp);
     pPsb->exaScratchSize = tmp * 1024;
 
-    pPsb->exaCached = TRUE;
-    from =
-	xf86GetOptValBool(pPsb->options, OPTION_EXACACHED, &pPsb->exaCached)
-	? X_CONFIG : X_DEFAULT;
-
-    if (!pPsb->noAccel)
-	xf86DrvMsg(pScrn->scrnIndex, from,
-		   "[EXA] Keep pixmap cache in %s memory.\n",
-		   pPsb->exaCached ? "cache coherent" : "uncached");
-
     return TRUE;
 }
 
@@ -743,6 +758,10 @@ psbInitOutputs(ScrnInfoPtr pScrn)
 
     xf86GetOptValBool(pPsb->options, OPTION_NOFITTING, &pPsb->noFitting);
 
+    xf86GetOptValBool(pPsb->options, OPTION_DOWNSCALE, &pPsb->downScale);
+
+    xf86GetOptValBool(pPsb->options, OPTION_VSYNC, &pPsb->vsync);
+
     xf86CrtcConfigInit(pScrn, &psbXf86crtcConfigFuncs);
     if (pScrn == pDevice->pScrns[0]) {
 	(void)psbAddToOutputList(pPsb, psbLVDSInit(pScrn, "LVDS0"));
@@ -990,6 +1009,10 @@ psbPreInit(ScrnInfoPtr pScrn, int flags)
 static Bool
 psbDeviceInit(PsbDevicePtr pDevice, int scrnIndex)
 {
+    #if (XORG_VERSION_MINOR >= 5 || (XORG_VERSION_MINOR == 4 && XORG_VERSION_PATCH == 99))
+        struct pci_device dev;
+    #endif
+
     PSB_DEBUG(scrnIndex, 3, "psbDeviceScreenInit\n");
 
     if (pDevice->refCount++ != 0) {
@@ -1012,9 +1035,14 @@ psbDeviceInit(PsbDevicePtr pDevice, int scrnIndex)
 
     if (!vgaHWGetHWRec(pDevice->pScrns[0]))
 	return FALSE;
-
-    pDevice->regPhys = pDevice->pciInfo->memBase[0];
-    pDevice->regSize = 1 << pDevice->pciInfo->size[0];
+  
+   #if (XORG_VERSION_MINOR >= 5 || (XORG_VERSION_MINOR == 4 && XORG_VERSION_PATCH == 99))
+        pDevice->regPhys = pDevice->pciInfo->regions[0].base_addr;
+        pDevice->regSize = pDevice->pciInfo->regions[0].size;
+   #else
+        pDevice->regPhys = pDevice->pciInfo->memBase[0];
+        pDevice->regSize = 1 << pDevice->pciInfo->size[0];
+   #endif    
 
     pDevice->regMap = xf86MapVidMem(scrnIndex, VIDMEM_MMIO_32BIT,
 				    pDevice->regPhys, pDevice->regSize);
@@ -1031,6 +1059,7 @@ psbDeviceInit(PsbDevicePtr pDevice, int scrnIndex)
 	       "\twith size %lu kiB\n", pDevice->regPhys,
 	       pDevice->regSize / 1024);
 
+
     /*
      * Map the OpRegion SCI region
      */
@@ -1040,7 +1069,16 @@ psbDeviceInit(PsbDevicePtr pDevice, int scrnIndex)
 	OpRegionPtr OpRegion;
 	char *OpRegion_String = "IntelGraphicsMem";
 
-	OpRegion_Phys = pciReadLong(pDevice->pciTag, 0xFC);
+
+        #if (XORG_VERSION_MINOR >= 5 || (XORG_VERSION_MINOR == 4 && XORG_VERSION_PATCH == 99))
+            dev.domain = PCI_DOM_FROM_TAG(pDevice->pciTag);
+            dev.bus = PCI_BUS_FROM_TAG(pDevice->pciTag);
+            dev.dev = PCI_DEV_FROM_TAG(pDevice->pciTag);
+            dev.func = PCI_FUNC_FROM_TAG(pDevice->pciTag);
+            pci_device_cfg_read_u32(&dev,&OpRegion_Phys,0xFC);
+        #else
+            OpRegion_Phys = pciReadLong(pDevice->pciTag, 0xFC); 
+        #endif
 
 	pDevice->OpRegion = xf86MapVidMem(scrnIndex, VIDMEM_MMIO_32BIT,
 					  OpRegion_Phys, OpRegion_Size);
@@ -1065,19 +1103,34 @@ psbDeviceInit(PsbDevicePtr pDevice, int scrnIndex)
 	    pDevice->OpRegion = NULL;
 	}
     }
+   
+    #if (XORG_VERSION_MINOR >= 5 || (XORG_VERSION_MINOR == 4 && XORG_VERSION_PATCH == 99))
+    struct pci_device device;
+    uint32_t temp;
+ 
+    device.domain = PCI_DOM_FROM_TAG(pDevice->pciTag);
+    device.bus = PCI_BUS_FROM_TAG(pDevice->pciTag);
+    device.dev = PCI_DEV_FROM_TAG(pDevice->pciTag);
+    device.func = PCI_FUNC_FROM_TAG(pDevice->pciTag);
+    pci_device_cfg_read_u32(&device,&temp, PSB_BSM);
+ 
+    pDevice->stolenBase = (unsigned long)temp & 0xFFFFF000;
+    xf86DrvMsg(scrnIndex,X_ERROR,"the stolenBase is:0x%08x\n",pDevice->stolenBase);
+    #else
+     pDevice->stolenBase = 
+       (unsigned long)pciReadLong(pDevice->pciTag, PSB_BSM) & 0xFFFFF000;
+    #endif
 
-    pDevice->stolenBase =
-	(unsigned long)pciReadLong(pDevice->pciTag, PSB_BSM) & 0xFFFFF000;
-    pDevice->stolenSize = PSB_READ32(PSB_PGETBL_CTL) & 0xFFFFF000;
-
-    pDevice->stolenSize -= pDevice->stolenBase;
-    pDevice->stolenSize -= PSB_BIOS_POPUP_SIZE;
-    pDevice->stolenSize >>= PSB_PAGE_SHIFT;
-
-    xf86DrvMsg(scrnIndex, X_PROBED,
+     pDevice->stolenSize = PSB_READ32(PSB_PGETBL_CTL) & 0xFFFFF000;
+     pDevice->stolenSize -= pDevice->stolenBase;
+     pDevice->stolenSize -= PSB_BIOS_POPUP_SIZE;
+     pDevice->stolenSize >>= PSB_PAGE_SHIFT;
+  
+     xf86DrvMsg(scrnIndex, X_PROBED,
 	       "Detected %lu kiB of \"stolen\" memory set aside as video RAM.\n",
 	       pDevice->stolenSize << 2);
 
+     
 #undef PSB_USE_MMU
 #undef PSB_WRITECOMBINING
 
@@ -1094,8 +1147,11 @@ psbDeviceInit(PsbDevicePtr pDevice, int scrnIndex)
 #endif /* PSB_WRITECOMBINING */
 
 #else
-    pDevice->fbPhys = pDevice->stolenBase;
-    pDevice->fbSize = pDevice->stolenSize << PSB_PAGE_SHIFT;
+   
+     pDevice->fbPhys = pDevice->stolenBase;
+     pDevice->fbSize = pDevice->stolenSize << PSB_PAGE_SHIFT;
+    
+    xf86DrvMsg(scrnIndex,X_ERROR,"screnIndex is:%d;fbPhys is:0x%08x; fbsize is:0x%08x\n",scrnIndex,pDevice->fbPhys,pDevice->fbSize);
     pDevice->fbMap = xf86MapVidMem(scrnIndex, VIDMEM_MMIO_32BIT,
 				   pDevice->fbPhys, pDevice->fbSize);
 #endif /* PSB_USE_MMU */
@@ -1165,9 +1221,21 @@ psbDeviceInit(PsbDevicePtr pDevice, int scrnIndex)
 	PCITAG host = pciTag(0, 0, 0);
 	CARD32 clock, period;
 
+        #if (XORG_VERSION_MINOR >= 5 || (XORG_VERSION_MINOR == 4 && XORG_VERSION_PATCH == 99))
+        struct pci_device dev;
+        dev.domain = PCI_DOM_FROM_TAG(host);
+        dev.bus = PCI_BUS_FROM_TAG(host);
+        dev.dev = PCI_DEV_FROM_TAG(host);
+        dev.func = PCI_FUNC_FROM_TAG(host);
+ 
+        pci_device_cfg_write_u32(&dev, 0xD0050300, 0xD0);
+        pci_device_cfg_read_u32(&dev, &clock, 0xD4);
+        #else
+        pciWriteLong(host, 0xD4, 0x00C32004);
+	pciWriteLong(host, 0xD0, 0xE0033000);
 	pciWriteLong(host, 0xD0, 0xD0050300);
-
 	clock = pciReadLong(host, 0xD4);
+        #endif
 
 	pDevice->CoreClock = CoreClocks[clock & 0x07];
 	pDevice->MemClock = MemClocks[(clock & 0x08) >> 3];
@@ -1199,9 +1267,21 @@ psbDeviceInit(PsbDevicePtr pDevice, int scrnIndex)
 //To read PBO.MSR.CCF Mode and Status Register C-Spec -p112
 	{
 	PCITAG host = pciTag(0, 0, 0);
-	
-	pciWriteLong(host, 0xD0, PCI_PORT5_REG80_FFUSE);
-	pDevice->sku_value= pciReadLong(host, 0xD4);
+
+        #if (XORG_VERSION_MINOR >= 5 || (XORG_VERSION_MINOR == 4 && XORG_VERSION_PATCH == 99))
+          struct pci_device dev;
+          dev.domain = PCI_DOM_FROM_TAG(host);
+          dev.bus = PCI_BUS_FROM_TAG(host);
+          dev.dev = PCI_DEV_FROM_TAG(host);
+          dev.func = PCI_FUNC_FROM_TAG(host);
+ 
+          pci_device_cfg_write_u32(&dev, PCI_PORT5_REG80_FFUSE, 0xD0);
+          pci_device_cfg_read_u32(&dev, &(pDevice->sku_value), 0xD4);
+        #else
+          pciWriteLong(host, 0xD0, PCI_PORT5_REG80_FFUSE);
+	  pDevice->sku_value= pciReadLong(host, 0xD4);
+        #endif
+
 	pDevice->sku_bSDVOEnable = (pDevice->sku_value & PCI_PORT5_REG80_SDVO_DISABLE)?FALSE : TRUE;
 	pDevice->sku_bMaxResEnableInt = (pDevice->sku_value & PCI_PORT5_REG80_MAXRES_INT_EN)?TRUE: FALSE;
 	xf86DrvMsg(scrnIndex, X_INFO,
@@ -1662,6 +1742,9 @@ psbEnterVT(int scrnIndex, int flags)
     if (!xf86SetDesiredModes(pScrn))
 	return FALSE;
     psbDescribeOutputConfiguration(pScrn);
+
+    PSB_WRITE32(DSPARB, (0x7E<<7 | 0x40<<0));
+
     psbAdjustFrame(pScrn->scrnIndex, pScrn->frameX0, pScrn->frameY0, 0);
     psbDRIUnlock(pScrn);
     return TRUE;
